import os
import uuid
import json
from pprint import pprint
import shutil
from datetime import date

from jinja2 import Environment, select_autoescape

from madoc.loader import MadocLoader

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DIR = os.getcwd()

env = Environment(
    loader=MadocLoader(os.path.join(BASE_DIR, "templates")),
    autoescape=select_autoescape()
)


def build_html(path, pages, links, dist_dir, level, bg_color, no_mark, title):
    """Convert all markdown files to html files"""
    template = env.get_template("madoc/recursive_render.html")
    pages = []
    uuid_value = str(uuid.uuid4())
    for file in os.listdir(path):
        page = {}
        if file.endswith(".md"):
            with open(os.path.join(path, file), "r") as md_file:
                page['content'] = md_file.read().replace("<", f"<{uuid_value}").replace(">", f"{uuid_value}>")
                page['name'] = file.replace(".md", "")
                pages.append(page)
    json_pages = json.dumps(pages)

    context = {
        "uuid": uuid_value,
        "index_link": level*"../" + "index.html",
        "favicon_url": level*"../" + "favicon.ico",
        "links": links,
        "bg_color": bg_color,
        "no_mark": no_mark,
        "title": title,
        "pages": json_pages,
    }

    with open(os.path.join(dist_dir, "documentation.madoc.html"), "w") as f:
        f.write(template.render(**context))


def parser(directory=DIR, bg_color="#fbfbfb", no_mark=False, include=None, dist_dir="madoc_dist", level=0):
    datas = { 'id': str(uuid.uuid4()), 'directory': directory, 'subdirs': [], 'files': [], 'foldable': True}
    is_valid = False
    links = []
    types_to_include = [
            # "html", "css", "js", "png", "jpg", "jpeg", "gif", "svg", "ico", "json", "pdf"
            ]
    if include is not None and include != 'ALL':
        types_to_include = include.split(",") + types_to_include
    for name in os.listdir(directory):
        if name.startswith(".") or name == 'madoc_dist':
            continue

        path = os.path.join(directory, name)
        # Directories
        if os.path.isdir(path):
            rec_datas, rec_is_valid = parser(directory=path, bg_color=bg_color, no_mark=no_mark, include=include, dist_dir=os.path.join(dist_dir, name), level=level+1)
            if rec_is_valid:
                links = [{
                    'name': link['directory'].split("/")[-1],
                    'url':"/".join(link['directory'].split("/")[-(rec_datas['level']-level):]) + "/documentation.madoc.html"
                    } for link in rec_datas['subdirs']]
                datas['subdirs'].append(rec_datas)
                # pprint(datas)
                # makes the tree in madoc_dist
                if not os.path.exists(os.path.join(dist_dir, name)):
                    os.makedirs(os.path.join(dist_dir, name))
                # sort rec_datas['files'] by name
                rec_datas['files'] = sorted(rec_datas['files'])
                if not rec_datas['files']:
                    rec_datas['foldable'] = False
                build_html(
                    path=path,
                    pages=rec_datas['files'],
                    links=links,
                    dist_dir=os.path.join(dist_dir, name),
                    level=level+1,
                    bg_color=bg_color,
                    no_mark=no_mark,
                    title=name,
                )
                is_valid = True
        # files
        elif name.lower().endswith(".md"):
            is_valid = True
            datas['files'].append(name)
        # copy or ignore files
        elif name.endswith(".madoc.html"):
            continue
        elif include == 'ALL' or name.split(".")[-1].lower() in types_to_include:
            if not os.path.exists(os.path.join(dist_dir)):
                os.makedirs(os.path.join(dist_dir))
            shutil.copyfile(path, os.path.join(dist_dir, name))
        datas['level'] = level
    return datas, is_valid


def index_builder(datas, bg_color="#fbfbfb", no_mark=False, title=""):
    date_created = date.today().strftime("%B %d, %Y")
    template = env.get_template("madoc/recursive_index.html")
    datas = [datas]
    # pprint(datas)
    context = {
        "bg_color": bg_color,
        "no_mark": no_mark,
        "datas": datas,
        "base_dir": "",
        "date_created": date_created,
        "title": title,
    }

    with open(os.path.join(DIR, 'madoc_dist', "index.html"), "w") as f:
        f.write(template.render(**context))


def main_recursive(
    bg_color="#fbfbfb",
    no_mark=False,
    title="",
    include=None
):
    # delete existant madoc_dist folder
    if os.path.exists(os.path.join(DIR, "madoc_dist")):
        shutil.rmtree(os.path.join(DIR, "madoc_dist"))

    # recreate an empty madoc_dist folder
    os.makedirs(os.path.join(DIR, "madoc_dist"))

    # Initialize the loop
    rec_datas, rec_is_valid = parser(directory=DIR, bg_color=bg_color, no_mark=no_mark, include=include,
                                     dist_dir=os.path.join(DIR, "madoc_dist"))

    if not rec_datas['files']:
        rec_datas['foldable'] = False

    if rec_is_valid:
        links = [{
            'name': link['directory'].split("/")[-1],
            'url':link['directory'].split("/")[-1] + "/documentation.madoc.html"
            } for link in rec_datas['subdirs']]
        datas = {'subdirs': [], 'files': []}
        datas['subdirs'].append(rec_datas)
        datas['id'] = uuid.uuid4()
        datas['foldable'] = True
        rec_datas['files'] = sorted(rec_datas['files'])
        build_html(
            path=DIR,
            pages=rec_datas['files'],
            links=links,
            dist_dir=os.path.join(DIR, "madoc_dist"),
            level=0,
            bg_color=bg_color,
            no_mark=no_mark,
            title=title,
        )
    index_builder(rec_datas, bg_color=bg_color, no_mark=no_mark, title=title)

    # copy the default favicon if none exists
    if not os.path.exists(os.path.join(DIR, "favicon.ico")):
        shutil.copy(
            os.path.join(BASE_DIR, "assets", "favicon.ico"),
            os.path.join(DIR, 'madoc_dist', "favicon.ico")
            )
    print("Madoc SUCCESS: recursive build done ! Check the 'madoc_dist' folder.")



# #! /usr/bin/env python3

# import base64, re
# import mimetypes
# from typing import Optional

# from madoc.silly_engine.text_tools import c
# from madoc.silly_engine.logger import Logger

# logger = Logger("base64 converter")
# image_pattern = r"!\[(.*?)\]\((.*?)\)"


# def encode_image_to_base64(image_path: str) -> Optional[str]:
#     """Encodes an image file to a base64 string with the correct MIME type."""
#     mime_type, _ = mimetypes.guess_type(image_path.strip())
#     if not mime_type:
#         mime_type = "application/octet-stream"  # Default if type is unknown

#     try:
#         with open(image_path.strip(), "rb") as image_file:
#             base64_string = base64.b64encode(image_file.read()).decode("utf-8")
#             return f"data:{mime_type};base64,{base64_string}"
#     except FileNotFoundError:
#         logger.warning(f"the file '{image_path}' was not found, it was then left as it is (not converted to base64).")
#         return None  # Return None if file is not found

# def replace_images_with_base64(text: str) -> str:
#     """Replaces all image file paths with their Base64 encoding, but keeps original if not found."""
#     pattern: str = r"!\[(.*?)\]\((.*?)\)"  # Matches ![image_name](image_file)

#     def replacer(match: re.Match) -> str:
#         image_name: str
#         image_file: str
#         image_name, image_file = match.groups()

#         base64_string: Optional[str] = encode_image_to_base64(image_file)
#         if base64_string:
#             return f"![{image_name}]({base64_string})"
#         else:
#             return match.group(0)  # Keep the original text if the file is not found

#     return re.sub(pattern, replacer, text)


import markdown
from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor
import re
import base64
import mimetypes
import requests
from typing import Optional


def encode_image_to_base64(image_path: str) -> Optional[str]:
    """Encodes an image file (local or web) to a Base64 string with the correct MIME type."""
    try:
        if image_path.startswith(("http://", "https://")):
            # Download the image from the web
            response = requests.get(image_path, timeout=5)
            if response.status_code == 200:
                mime_type = response.headers.get("Content-Type", "application/octet-stream")
                base64_string = base64.b64encode(response.content).decode("utf-8")
                return f"data:{mime_type};base64,{base64_string}"
            else:
                return None  # Keep the original URL if the download fails
        else:
            # Handle local file
            mime_type, _ = mimetypes.guess_type(image_path)
            if not mime_type:
                mime_type = "application/octet-stream"

            with open(image_path, "rb") as image_file:
                base64_string = base64.b64encode(image_file.read()).decode("utf-8")
                return f"data:{mime_type};base64,{base64_string}"
    except (FileNotFoundError, requests.RequestException):
        return None  # Return None if file not found or request fails

class ImageBase64Processor(Preprocessor):
    """Custom Preprocessor to replace image paths with Base64, but ignore code blocks."""

    def __init__(self, md):
        super().__init__(md)
        self.in_code_block = False
        self.image_md_pattern = re.compile(r"!\s*\[\s*(.*?)\s*\]\s*\(\s*(.*?)\s*\)")  # Matches ![alt](image)
        self.image_html_pattern = re.compile(r'<img\s+[^>]*?\bsrc\s*=\s*["\'](.*?)["\'][^>]*?/?>', re.IGNORECASE)  # Matches <img src="image">

    def run(self, lines):
        processed_lines = []

        for line in lines:
            # Toggle in_code_block when encountering ```
            if line.strip().startswith(("```", "~~~")):
                self.in_code_block = not self.in_code_block
                processed_lines.append(line)
                continue

            if not self.in_code_block:
                # Replace Markdown images
                line = self.image_md_pattern.sub(self.replace_md_image, line)

                # Replace HTML images
                line = self.image_html_pattern.sub(self.replace_html_image, line)

            processed_lines.append(line)

        return processed_lines

    def replace_md_image(self, match):
        """Replaces Markdown image paths with Base64."""
        alt_text, image_path = match.groups()
        base64_str = encode_image_to_base64(image_path)
        return f"![{alt_text}]({base64_str})" if base64_str else match.group(0)

    def replace_html_image(self, match):
        """Replaces HTML image paths with Base64."""
        image_path = match.group(1)
        base64_str = encode_image_to_base64(image_path)
        return match.group(0).replace(image_path, base64_str) if base64_str else match.group(0)

class ImageBase64Extension(Extension):
    """Markdown extension to replace images with Base64 outside of code blocks."""

    def extendMarkdown(self, md):
        md.preprocessors.register(ImageBase64Processor(md), 'image_base64_processor', 25)

def convert_images_to_base64(markdown_text: str) -> str:
    """Replaces Markdown and HTML images with Base64 encoding, but ignores images inside code blocks."""
    md = markdown.Markdown(extensions=[ImageBase64Extension()])
    return md.convert(markdown_text)

# Example Markdown with local & web images
markdown_text = """
# Example Markdown

Local image (should convert):
![Local](images/picture.jpg)

Web image (should convert if downloadable):
![Web Image](https://example.com/image.png)

HTML image from the web:
<img src="https://example.com/photo.jpg" alt="Photo">
"""


#! /usr/bin/env python3
# coding: utf-8

import os
import json
import uuid
import argparse

from jinja2 import Environment, select_autoescape
import markdown

from madoc.loader import MadocLoader
from madoc import __version__
from madoc.madoc_recursive import main_recursive
from madoc.base64_converter import convert_images_to_base64

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DIR = os.getcwd()

# defaults
no_b64 = False

env = Environment(
    loader=MadocLoader(os.path.join(BASE_DIR, "templates")),
    autoescape=select_autoescape()
)

template = env.get_template("madoc/render.html")


def cmd(*args, **kwargs) -> None:
    """Command line interface"""
    # positionnal arguments
    parser = argparse.ArgumentParser(
        description=(
            "Convert all markdown files in the directory into a single html file. "
            "Or recursively gather your documentation in an index.html file and subfolders, usable as a static site."
            )
        )
    parser.add_argument("-b", "--bg-color", type=str, default="#fbfbfb", help="Background color, default is #fbfbfb")
    parser.add_argument("-t", "--title", type=str, default="Documents", help="Title of the page, default is 'Documents' (root page if recursive)")
    # flags (store_true)
    parser.add_argument("-V", "--version", help="Version", action="store_true")
    parser.add_argument("-c", "--code", help="enable code coloring", action="store_true")
    parser.add_argument("-r", "--recursive", help="Recursive action, links the subfolders to an index", action="store_true")
    parser.add_argument("--no-mark", help="remove 'generated by Madoc' mark", action="store_true")
    parser.add_argument("--no-b64", help="do not convert images to base64", action="store_true")
    parser.add_argument("-i", "--include", help="Included files types in recursive mode, e.g: -i ALL | -i jpg,jpeg,png", default=None)
    args = parser.parse_args()
    no_mark = False
    if args.version:
        print(__version__)
        return
    if args.code:
        print("= syntax coloring enabled")
    if args.no_mark:
        no_mark = True
    if args.no_b64:
        no_b64 = True
        print("= no b64 images")
    if args.recursive:
        main_recursive(bg_color=args.bg_color, no_mark=no_mark, title=args.title, include=args.include)
        return
    main(title=args.title, syntax_color=args.code)


def main(
    title="Documents",
    syntax_color=False
) -> None:
    """Convert all markdown files to html files"""
    pages = []
    files = sorted(os.listdir(DIR))
    for file in files:
        page = {}
        if file.endswith(".md") or file.endswith(".MD"):
            print(file)
            with open(file, "r") as md_file:
                page['name'] = file.replace(".md", "").replace(".MD", "")
                raw_content = md_file.read()
                extentions = [
                    "abbr",              # Abbreviations
                    "admonition",        # Notes, Warnings, etc.
                    "attr_list",         # HTML attributes on elements
                    "def_list",          # Definition lists
                    "fenced_code",       # Triple-backtick code blocks
                    "footnotes",         # Footnotes
                    "legacy_attrs",      # Support for legacy attributes
                    "legacy_em",         # Legacy emphasis handling
                    "md_in_html",        # Markdown inside raw HTML blocks
                    "meta",              # YAML-style metadata
                    "nl2br",             # Newline to <br> conversion
                    "sane_lists",        # Better list handling
                    "smarty",            # Smart punctuation (quotes, dashes)
                    "tables",            # GitHub-style tables
                    "toc",               # Table of Contents
                    "wikilinks",         # Wiki-style links])
                ]
                if syntax_color:
                    extentions.append("codehilite")  # Syntax highlighting (requires Pygments))
                page["content"] = markdown.markdown(raw_content, extensions=extentions)
                if not no_b64:
                    page["content"] = convert_images_to_base64(page["content"])
            pages.append(page)

    context = {
        "title": title,
        "pages": pages,
        "css_style": "madoc/madoc_style.css"
    }

    with open("documentation.madoc.html", "w") as f:
        f.write(template.render(**context))

    print("Done !")


if __name__ == "__main__":
    cmd()



"""
This modules provides a quick and easy way to use Logger class.
example:

logger = Logger("My Application", level="DEBUG")

logger.debug("that works")

"""


import logging

levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

class CustomFormatter(logging.Formatter):
    blue = "\033[34m"
    green = "\033[32m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    prefix = "[%(asctime)s][%(name)s][%(levelname)s] "
    message = "%(message)s"

    FORMATS = {
        logging.DEBUG: green + prefix + reset + message,
        logging.INFO: blue + prefix + reset + message,
        logging.WARNING: yellow + prefix + reset + message,
        logging.ERROR: red + prefix + reset + message,
        logging.CRITICAL: bold_red + prefix + reset + message
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

class Logger(logging.Logger):
    def __init__(self, name :str=__name__, level: str="INFO"):
        super().__init__(name)
        self.console_handler = logging.StreamHandler()
        self.console_handler.setFormatter(CustomFormatter())
        self.addHandler(self.console_handler)

    def setLevel(self, level):
        super().setLevel(levels[level])
